/*
; Copyright (c) 1998-2011 Mikhail Ranish
;
; Permission to use, copy, modify, and/or distribute this software for any
; purpose with or without fee is hereby granted, provided that the above
; copyright notice and this permission notice appear in all copies.
;
; THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
; WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
; MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
; ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
; WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
; ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
; OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

//#define VERBOSE
//#define MYRUNTIME (char*)0x400000
#define GEN_FAT
#define GEN_ISO
//#define GEN_ZIP

#include "fat.h"

#ifdef GEN_FAT
#include <windows.h>
#endif

typedef FILE* OutStream;
#define WriteStream(ff,p,n) fwrite(p,1,n,ff)

#define BOOTLABEL "LOADER  EXE"
#define DOS_DATE(y,m,d) (((y)-1980)<<9|(m)<<5|(d))
#define DOS_TIME(h,m,s) ((h)<<11|(m)<<5|(s)>>1)
#define PM_DATE  DOS_DATE(2011,7,4)
#define PM_TIME  DOS_TIME(01,02,00)

int create_144_image(OutStream ff, char* bootFat, char* bootExe, char *label, int bootExeSize);
int create_iso_image(OutStream ff, char* bootFat, char* bootExe, char *label, int bootExeSize, int ISOmode);

void installBootSector(char *FATvol, char*buf3, char* label);


const char *cannotOpen = "Cannot open %s\n";


#ifdef VERBOSE
	int debug_mode=0;
#endif

struct AccState
{
	char *pdst;
	long *pacc;
	int   accb;
	int	  prev_op;
	unsigned  n_prefix; // 1..9
	unsigned  n_fixed;  // 1..9
	unsigned  n_extra;  // 2..14 // even values only
};

void encode(struct AccState* pAS, int op, int len, char *psrc)
{
	if( op==0 ) // init
	{
		pAS->pdst=psrc;
		*(unsigned long*)(pAS->pdst) =
			((unsigned)1L << (32-pAS->n_fixed))+
						256*((16*pAS->n_extra+4)-pAS->n_prefix)+
							(256-pAS->n_prefix);
		pAS->pdst+=4;
		pAS->pacc=(long*)(pAS->pdst);
		*(pAS->pacc)=0;
		pAS->pdst+=4;
		pAS->accb=0;
		pAS->prev_op=0;
		return;
	}

	int x = (len<<1);
	if( op==1 ) // move
	{
		x++;
	}
	else if( op==2 ) // copy
	{
		x++;
	}
	else if( op==3 ) // push / exit / base
	{
//	#ifdef VERBOSE
//		if( debug_mode )
//			printf("(len=%d) ", len );
//	#endif
	}

	if( pAS->prev_op!=1 && op!=1 ) // was not move and is not move
	{
		x = (x<<1);
	}
	pAS->prev_op=op;
	// encode x
	// ASSERT(x<(1<<22)
	int xn; // bits
	long z = (1<<pAS->n_fixed);
	long zsh1 = (z<<pAS->n_extra);
	long zsh2 = (zsh1<<pAS->n_extra);
	if( pAS->n_prefix>1 && x<z*(pAS->n_prefix-1) )
	{
		xn = (x/z)+1+pAS->n_fixed;
		x = z + x%z;
	}
	else if( x<zsh1 )
	{
		xn=pAS->n_prefix + pAS->n_fixed + pAS->n_extra;
		//x|=zsh1;
	}
	else if( x<zsh2 )
	{
		xn=pAS->n_prefix + pAS->n_fixed + 2*pAS->n_extra;
		x|=zsh2;
	}
	else
	{
		printf("\nError: need larger -X pfe\n");
		exit(1);
	}
//	printf(" {op=%d,len=%x,x=%x,n=%d}\n",op,len,x,xn);
	#ifdef VERBOSEz
		if( debug_mode )
			printf("(n=%d) ", xn );
	#endif

	//push code into control stream
	while(xn>0)
	{
		if( pAS->accb==32 )
		{
			pAS->pacc = (long*)(pAS->pdst);
			*(pAS->pacc)=0;
			pAS->accb=0;
			pAS->pdst+=4;
		}
		xn--;
		if( xn<=31 && x&(1<<xn) ) // take high bit -> push it on the bottom of acc
			*(pAS->pacc) |= (1<<pAS->accb);
		pAS->accb++;
	}
	if( op==1 ) // move len+1 bytes
	{
		int i;
		for( i=0 ; i<=len ; i++ )
		{
			*(pAS->pdst++)=*(psrc++);
		}

	}
}

unsigned hash(char*p) // a variation of djb2 bernstein hash
{
	int h=5381;//=0x1505
	do{ h=(h*33)^(*p); }while(*p++);
	return h;
}

struct pe_sect
{
	char name[8];
	long virt_size;
	long virt_addr;
	long file_size;
	long file_addr;
	long ptr_to_reloc;
	long ptr_to_linenums;
	short num_reloc;
	short num_linenums;
	long sect_characteristics;
};

unsigned readFile(char* fileName,char**pbuf)
{
	unsigned res=0;
	FILE *fin = fopen(fileName,"rb");
	if( fin==NULL )
	{
		printf(cannotOpen,fileName);
		exit(1);
	}
	fseek(fin,0,SEEK_END);
	res = ftell(fin);
	fseek(fin,0,SEEK_SET);
	int size = (res+1023)&(~1023);
	*pbuf=(char*)realloc(*pbuf,size);
	//memset(*pbuf,0,size);
	fread(*pbuf,1,res,fin);
	fclose(fin);
	return res;
}

FILE *fopenWB(char *fileName)
{
	FILE *fout = fopen(fileName,"wb");
	if( fout==NULL )
	{
		printf(cannotOpen,fileName);
		exit(1);
	}
	return fout;
}


int main(int argc, char**argv, char**env)
{
	int i, j, k, m;
	long flen;
	char *buf=0;
	char *infile=0;
	char *outfile=0;
	char *FLPfile=0;
	char *ISOfile=0;
	int ISOmode=0;
	char *bootfile=0;
	char *stubfile=0;
	char *plabel=BOOTLABEL;
	char *FATvol=0;
	struct AccState as;
	as.n_prefix = 3;
	as.n_fixed  = 5;
	as.n_extra  = 8;

	int bAddZip = 0;

	for( i=0 ; i<argc ; i++ )
	{
		if( argv[i][0]=='-' )
		{
			char opt=argv[i][1];

#ifdef GEN_ZIP
			if( opt=='z' )
			{
				bAddZip = 1;
				continue;
			}
#endif
#ifdef VERBOSE
			if( opt=='v' )
			{
				debug_mode=1;
				continue;
			}
#endif
			if( i+1>=argc ) break;
			long lopt=*(long*)(argv[i]);
			char *parg = argv[++i];
			if( opt=='i' )
			{
				infile=parg;
			}
			else if( opt=='b' )
			{
				bootfile=parg;
			}
			else if( opt=='L'  )
			{
				plabel=parg;
			}
			else if( opt=='s' )
			{
				stubfile=parg;
			}
			else if( opt=='o' )
			{
				outfile=parg;
			}
			else if( lopt==0x504C462D ) // -FLP
			{
				FLPfile=parg;
			}
#ifdef GEN_ISO
			else if( lopt==0x3044432D ) // -CD0
			{
				ISOfile=parg;
			}
			else if( lopt==0x3244432D ) // -CD2
			{
				ISOmode=2;
				ISOfile=parg;
			}
#endif
#ifdef GEN_FAT
			else if( lopt==0x5441462D ) // -FAT
			{
				FATvol=parg;
			}
#endif
			else if( opt=='X' )
			{
				as.n_prefix = parg[0]-'0';
				as.n_fixed  = parg[1]-'0';
				as.n_extra  = atoi(parg+2)&(~1); // even values only
			}
		}
	}


	if( infile==0 && FATvol==0
		|| stubfile==0 && outfile!=0
		|| bootfile==0 &&
			( FLPfile!=0
#ifdef GEN_ISO
			|| ISOfile!=0
#endif
			|| FATvol!=0
			)
#ifdef MYRUNTIME
//		&& *(long*)(MYRUNTIME+0x28)!=0x726D010b
#endif
	)
	{
		printf(
			"Usage: pe2boot [-X 112..9914] -s pe_stub.bin	-o out.exe -i in.exe	<- compress PE\n"
			"   or: pe2boot -b boot.bin -L label -FLP"
#ifdef GEN_ISO
			"|-CD0|-CD2"
#endif
														 " out.img -i in.exe	<- make floppy/CD image\n"
			"   or: pe2boot -b boot.bin -L label -FAT \\\\.\\U:				<- make floopy/USB bootable\n");
		return 1;
	}

if( infile )
{
	flen = readFile(infile,&buf);
	if( flen==0 )
	{
		return 1;
	}

	if( stubfile!=0 )
	{
		long pe_off = *(long*)(buf+0x3C);
		char *pe = buf+pe_off;
		if( buf[0]!='M' || buf[1]!='Z' ||
			pe_off >= flen ||
			*(long*)pe!=0x4550 || // PE32 magic
			//pe[0x18]!=0xB || pe[0x19]!=1 ||
			(pe[0x16]&1)!=0 ) // Relocation info stripped!
		{
			printf("%s not relocatable\n",infile);
			return 1;
		}

		long num_sections	= *(short*)(pe+0x06);
	//	long time_date_stamp = *(long*)(pe+0x08);
		long opt_hdr_size	= *(short*)(pe+0x14);
		//long exe_options	= *(short*)(pe+0x16);
		long entry_point	 = *(long*)(pe+0x28);
		long pe_image_base	 = *(long*)(pe+0x34);
		long size_of_image	 = *(long*)(pe+0x50);
		long size_of_header	 = *(long*)(pe+0x54);
		long subsystem	 	= *(short*)(pe+0x5C);

		long num_rvas		 = *(long*)(pe+0x74);

		long import_dir_addr= *(long*)(pe+0x78 + 1*8 );
		long import_dir_size= *(long*)(pe+0x78 + 1*8 + 4 );

		long base_reloc_addr = *(long*)(pe+0x78 + 5*8 );
		long base_reloc_size = *(long*)(pe+0x78 + 5*8 + 4 );

		long debug_info_addr = *(long*)(pe+0x78 + 6*8 );
		long debug_info_size = *(long*)(pe+0x78 + 6*8 + 4 );


		struct pe_sect *sect = (struct pe_sect *) ( pe + 0x18 + opt_hdr_size );
		long buf2size = size_of_header + size_of_image + 4096;

		char* buf2;
		buf2 = (char*)malloc(buf2size);
	#ifdef VERBOSE
		if( debug_mode )
		{
			printf("base  = %04X\n", pe_image_base );
			printf("entry = %04X\n", entry_point);
			printf("vsize = %04X\n", buf2size);
			printf("reloc = %04X, %04X\n", base_reloc_addr, base_reloc_size);
			printf("debug = %04X, %04X\n", debug_info_addr, debug_info_size);
		}
	#endif
		memset(buf2,0,buf2size);

		long min_vaddr=0;
		long max_vaddr=0;

		char* rptr=buf2 + base_reloc_addr;
		long rsize=base_reloc_size;

		for( i=0 ; i<num_sections ; i++ )
		{
	#ifdef VERBOSE
			if( debug_mode )
				printf("%-08s  F = {%04X,%04X}  V = {%04X,%04X}\n",
				sect->name,
				sect->file_addr, sect->file_size,
				sect->virt_addr, sect->virt_size);
	#endif
			long sect_vtop = sect->virt_addr + sect->virt_size;

			memmove(buf2+sect->virt_addr, buf+sect->file_addr, sect->file_size);

			if( (sect->sect_characteristics & 0x02000000)==0 ) // Bit set if section is discardable
			{
				if( i==0 || min_vaddr > sect->virt_addr ) min_vaddr = sect->virt_addr;
				if( i==0 || max_vaddr < sect_vtop       ) max_vaddr = sect_vtop;
			}
			sect++;
		}
	#ifdef VERBOSE
		if( debug_mode )
		{
			printf("vmem=%X,%X\n", min_vaddr, max_vaddr);
		}
	#endif
		char temp_debug_info_dir[0x1C+0x40];
		if( debug_info_size > 0 )
		{
			memmove(temp_debug_info_dir,buf2+debug_info_addr,0x1C);
			memset(buf2+debug_info_addr,0,debug_info_size);

			debug_info_size = *(long*)(temp_debug_info_dir+0x10);
			debug_info_addr = *(long*)(temp_debug_info_dir+0x14);

			memmove(temp_debug_info_dir+0x1C,buf2+debug_info_addr, debug_info_size>0x40?0x40:debug_info_size);
			memset(buf2+debug_info_addr,0,debug_info_size);
		}

		int nbrr = 0;
		long* pbrr = 0;
		long pbrrSize = 0;
		long blockVA;
		long blockSize = 0;
		while(1)
		{
			if( nbrr>=pbrrSize )
			{
				pbrrSize+=64;
				pbrr=(long*)realloc(pbrr,pbrrSize*4+16);
			}
			if( blockSize==0 )
			{
				if( !rptr ) break;
				blockVA   = *(long*)rptr;
				blockSize = *(long*)(rptr+4) - 8;
				if( blockVA == 0 ) break;
				rptr+=8;
				rsize-=8;
	#ifdef VERBOSE
	//			if( debug_mode )
	//				printf("VA: %04X, %04X: ", blockVA, blockSize);
	#endif
			}
			if( rsize==0 ) break;

			long zz = *(short*)rptr;
			rptr+=2;
			rsize-=2;
			blockSize-=2;
			if( zz==0 ) continue;
			long rva = blockVA + (zz&0xFFF);
	#ifdef VERBOSE
			//	if( debug_mode )
			//		printf(" %04X (+%04X) ", rva, *(long*)(buf2+rva) );
	#endif
			zz = *(long*)(buf2+rva) -= pe_image_base + min_vaddr;
			if( zz<0 )
			{
				printf("Reloc<0 @ %X\n",rva);
				return 1;
			}
			pbrr[nbrr++] = rva - min_vaddr;
	#ifdef VERBOSE
				//if( debug_mode )
				//	printf("pbrr[%d]=%x\n",nbrr-1, rva-min_vaddr);
	#endif
		}


		long xsize = max_vaddr-min_vaddr;
		long xentry = entry_point - min_vaddr;


		int nd32;

	#ifdef MYRUNTIME
		if( stname==0 )
		{
			nd32=*(long*)(MYRUNTIME+0x68);
			memmove( buf, MYRUNTIME,nd32);
		}
		else
	#endif
		{
			char *buf3=0;
			nd32 = readFile(stubfile,&buf3);
			if( nd32==0 )
			{
				return 1;
			}
			memmove(buf,buf3,nd32);
			//free(buf3);
		}

		sect=0;
		pe_off=*(long*)(buf+0x3C);
		pe=buf+pe_off;
		if( buf[0]=='M' && buf[1]=='Z' &&
			pe_off<nd32
			// && *(long*)pe == 0x4550 // PE32 magic
			 &&	*(long*)(pe+0x18)==0x726D010b // runtime signature "mr"
			 //&&	*(long*)(buf+0x28)==0x001A010b // runtime signature
			 )
		{
			long opt_hdr_size	= *(short*)(pe+0x14);
			sect = (struct pe_sect *) ( pe + 0x18 + opt_hdr_size );
			nd32 += 4;
		}

		if( import_dir_addr )
		{
			unsigned *imp_ent, *psv;
			imp_ent = psv = (unsigned*)(buf2+import_dir_addr);
			for( ; imp_ent[3] ; imp_ent+=5 ) // cycle for each imported DLL
			{
				//imp_ent: hintTable, dateStamp, fwChain, dllName, thunkTable
				if( imp_ent[0] ) // hintTable
				{
					unsigned *htab=(unsigned*)(buf2+imp_ent[0]);
					while(*htab) *(htab++)=0; // erase hintTable
				}
				//imp_ent[0]=0; // hintTable
				//imp_ent[1]=0; // dateStamp
				//imp_ent[2]=0; // fwChain
				if( imp_ent[3] && imp_ent[4] )
				{
					*(psv++)=imp_ent[3] - min_vaddr; // save dllname
					*(psv++)=imp_ent[4] - min_vaddr; // save thunkTable
					char *dllname = buf2+imp_ent[3]; // dllName
					unsigned *ttab=(unsigned*)(buf2+imp_ent[4]); // thunkTable address
					if(  *(long*)(dllname)==0x4356534d &&
						(*(long*)(dllname+4)|0xF00)==0x2e303F52 ) // MSVCR?0.dll
					{
						(*(long*)(dllname+4))=0x2e2e5452; // Change dll to "MSVCRT"
					}
					while(*dllname&&*dllname!='.') dllname++;
					while(*dllname) *dllname++=0;
					while(*ttab) // thunkTab: replace list of pointers to function names with list of hashes
					{
						char* fname=buf2+*ttab;
						*(fname++)=0;
						*(fname++)=0;
						*ttab=hash(fname);
#ifdef VERBOSE
						if( debug_mode )
						{
							printf("hash = %08X for \"%s\"\n", *ttab, fname);
						}
#endif
						while(*fname) *(fname++)=0; // then erase function names
						ttab++;
					}
				}
				//imp_ent[3]=0; // dllName
				//imp_ent[4]=0; // thunkTable
			}
			import_dir_size = 4*(psv - (unsigned*)(buf2+import_dir_addr));
			while(psv!=imp_ent) *(psv++)=0;//clear remainder of the import table
			import_dir_addr -= min_vaddr;
		}
		else if( sect )
		{
			import_dir_addr=xsize;
			*(long*)(buf2+min_vaddr+xsize)=0;
			xsize+=4;
		}

		if( !sect )
		{
			import_dir_addr = -1;
			import_dir_size = 0;
		}

	#ifdef VERBOSE22
		//if( debug_mode )
		{
			FILE *fout = fopen("pe2bootzz20.out","wb");
			fwrite(buf2+min_vaddr,xsize,1,fout);
			fclose(fout);
		}
	#endif
	#ifdef VERBOSE
		if( debug_mode )
			printf("<xsize=%x>\n",xsize);
		int last_print=-100;
	#endif


		encode(&as,0,0,buf+nd32); // init
		i=0;
		j=0;
		k=0;
		m=0; // pending move m bytes
		int cstack=0;
		int xcopyi=0;
	#define MAXSTACK 16
		long xcopy[MAXSTACK];
		long xcopylen[MAXSTACK];
		char *pp = buf2+min_vaddr;
	//	Assuming pbrr[nbrr] are in ascending order
		pbrr[nbrr++]=xsize;
		pbrr[nbrr]=xsize+1;
		//char* cflags = (char*)malloc(xsize);
		//memset(cflags,0,xsize+1);
		//cflags[xentry]=254;
		//cflags[xsize]=255;
		while(i<xsize)
		{
			long nbrk = pbrr[k];
			long nbr = xsize;
			if( nbrk<nbr ) nbr = nbrk;
			if( xentry>=0 && xentry<nbr ) nbr = xentry;
			if( import_dir_addr>=0 && import_dir_addr<nbr ) nbr = import_dir_addr;
			if( xcopyi>0 && xcopy[xcopyi-1]<nbr ) nbr = xcopy[xcopyi-1];

			if( m!=0 && i==nbr )
			{
				encode(&as,1,m-1,pp+i-m);  // move m bytes
				m=0;
			}

#ifdef VERBOSE
		if( debug_mode )
		{
			if( (i%16==0) || i>last_print+24 )
			{
				printf("\n%04X: ",i);
				last_print=i;
			}
		}
	#endif

			if( i==xentry )
			{
	#ifdef VERBOSE

				if( debug_mode )
					printf("<entry> ");
	#endif
				encode(&as,3,cstack,0);	// push <entry> to the top of the stack
				cstack++;
				xentry=-1;
				continue;
			}

			if( i==import_dir_addr )
			{
	#ifdef VERBOSE

				if( debug_mode )
					printf("<impdir> ");
	#endif
				encode(&as,3,cstack-(xentry==-1?1:0),0);	// push <impdir> just below <entry>
				cstack++;
				while(import_dir_size)
				{
					int off = *(long*)(pp+i);
	#ifdef VERBOSE
				if( debug_mode )
					printf("<d=%x> ",off);
	#endif
					encode(&as,3,(cstack+1)+2*(off+1)+1,0); // base for dll name or ttab
					i+=4;
					import_dir_size-=4;
				}
				import_dir_addr=-1;
				//import_dir_size=0;
				continue;
			}

			if( xcopyi>0 && i==xcopy[xcopyi-1] )
			{
				if( m!=0 ) // move m bytes
				{
					encode(&as,1,m-1,pp+i-m);
					m=0;
				}
				xcopyi--;
				int nn=xcopylen[xcopyi];
	#ifdef VERBOSE
				if( debug_mode )
					printf(" <copy %04X> ", nn );
	#endif
				encode(&as,2,(nn)-2,0);// copy nn bytes
				cstack--;
				i+=nn;
				while( pbrr[k]<i ) k++;
				continue;
			}
			//if( cflags[i]!=0 && 0)
			{
				//if( m!=0 ) // move m bytes
				{
				//	encode(&as,1,m-1,pp+i-m);
				//	m=0;
				}
				//if( cflags[i]==1 )
				{
					//push addr
				}
				//else if( cflags[i]==2 ) // copy nn bytes
				{
					//int nn =
				}
			}

			if( i==nbrk )
			{
				int off = *(long*)(pp+i);
	#ifdef VERBOSE
				if( debug_mode )
					printf("b+%04X ", off );
	#endif
				encode(&as,3,(cstack+1)+2*(off+1)+1,0); // base
				i+=4;
				k++;
				continue;
			}
#if 1 // has bugs
			if( xcopyi<MAXSTACK && pp[i]!=pp[i+1] )
			{
				long maxn = 4;
				long maxp = 0;
				/**/long xlim = (xcopyi==0?xsize:xcopy[xcopyi-1]);
				for( j=i+2 ; j-i < 1500 && j<xlim ; j++ )
				{
					int nn=0;
					/**/while( j+nn<xlim && pp[i+nn]==pp[j+nn] ) nn++;
					//nnwhile( cflags[j+nn]==0 && pp[i+nn]==pp[j+nn] ) nn++;
					if( nn>maxn )
					{
						int kk=k;
						int je = j+nn;
						while( pbrr[kk]+4<=j ) kk++;
						if( pbrr[kk]>=je &&
							(xentry<=j || je<=xentry ) &&
							(import_dir_addr+import_dir_size<=j || je<=import_dir_addr) )
						{
							maxn=nn;
							maxp=j;
						}
						//if( nn>5 )
						//printf("\n%x -> %x, %x %s   pbrk[%d]=%x xentry=%x  impDir=%x,%x\n",i,j,nn,maxp==j?"*":"x", kk+1,pbrr[kk+1],xentry,import_dir_addr,import_dir_size);
					}
				}
				//if( maxn*maxn>=20+(j-i-75)/250 || maxn>=5 && (j-i)<150 ) // kind of works
				if( maxn>4 )
				{
					if( m!=0 ) // move m bytes
					{
						encode(&as,1,m-1,pp+i-m);
						m=0;
					}
					xcopy[xcopyi] = maxp;
					xcopylen[xcopyi] = maxn;
	#ifdef VERBOSE
				if( debug_mode )
					printf(" <addr @%x,+%x,%x> ", i,maxp-i,maxn );
	#endif
					encode(&as,3,0,0);	// push addr to the bottom of the stack
					cstack++;
					xcopyi++;
				}
			}

#endif

			long *pll = (long*)(pp+i);
			if( *pll >= -1 && *pll<(128<<as.n_fixed) && (pp[i+3]!=pp[i+4]) && i+3<nbr  )
			{
				if( m!=0 ) // move m bytes
				{
					encode(&as,1,m-1,pp+i-m);
					m=0;
				}
#ifdef VERBOSE
					if( debug_mode )
						printf("i=%02X ", *pll );
#endif
					encode(&as,3,(cstack+1)+2*(*pll+1)+0,0); // int32
					i+=4;
					continue;
			}

			for( j=i+1 ; j<nbr && pp[i]==pp[j] ; j++ );
			//if( m==0 && j-i>=3 || m<8 && j-i>=4 || j-i>=5 )
			//if( m<4 && j-i>=3 || m<16 && j-i>=4 || j-i>=5 )
			//if( (j-i-2)>m/8 )
			if( ((j-i-3)>m/20) )
			{
				if( m!=0 ) // move m bytes
				{
					encode(&as,1,m-1,pp+i-m);
					m=0;
				}

				encode(&as,3,0,0);		// push addr to the bottom of the stack

#ifdef VERBOSE
				if( debug_mode )
					printf("%02X*%04X ", (unsigned char)(pp[i]), (j-i) );
#endif

				if( j-i >= 5 && (pp[i]==0xFF || pp[i]==0) )
				{
#ifdef VERBOSE
					if( debug_mode )
						printf("z=%02X ", (unsigned char)(pp[i]) );
#endif
					encode(&as,3,(cstack+1/*cstack+1..-1*/+1)+2*(pp[i]+1)+0,0); // int32
					encode(&as,2,(j-i-4)-2,0);// copy n-4 bytes
				}
				else
				{
					encode(&as,1,0,pp+i);	// move 1 byte
					encode(&as,2,(j-i-1)-2,0);// copy n-1 bytes
				}

				i=j;
				continue;
			}
#ifdef VERBOSE
			if( debug_mode )
				printf("%02X ", (unsigned char)pp[i] );
#endif
			i++;
			m++;
		}
		if( m!=0 ) // move m bytes
		{
			encode(&as,1,m-1,pp+i-m);
			m=0;
		}
		encode(&as,3,(cstack+1)+2*(-1+1)+1,0);		// exit
#ifdef VERBOSE
		if( debug_mode )
			printf("\n");
#endif
		long csize = as.pdst-buf;
		if( (char*)as.pacc==as.pdst-4 )
			csize-=(32-as.accb)/8;

#if 0
		if( debug_info_size > 0 )
		{
			memset(buf+csize,0,0x1C+*(long*)(temp_debug_info_dir+0x10)+8);
			csize = (csize+3)&(~3);
	//		/*debug_info_size*/ *(long*)(temp_debug_info_dir+0x10) = 0x18+0xE;
			/*debug_info_addr*/ *(long*)(temp_debug_info_dir+0x14) = 0x1000 + csize + 0x1C;
			/*debug_info_raw */ *(long*)(temp_debug_info_dir+0x18) = csize+0x1C;
			memmove(buf+csize,temp_debug_info_dir,0x1C+*(long*)(temp_debug_info_dir+0x10));
			//memmove(buf+csize,temp_debug_info_dir,0x1C+0x18);
			//memmove(buf+csize+0x1C+0x18,"boot_test.pdb",0xE);

			*(long*)(buf+0x18) = time_date_stamp;
			*(long*)(buf+0x1C) = 0;
			*(long*)(buf+0x54) = 0;
			*(long*)(buf+0x26)&=~8;
			*(long*)(buf+0x84)=7;
			*(long*)(buf+0x8C)=0;
			*(long*)(buf+0x94)=0;

			*(long*)(buf+0xB8)=0x1000 + csize;
			*(long*)(buf+0xBC)=0x1C;
			//csize+=0x1C+0x18+0xE;
			csize+=0x1C+*(long*)(temp_debug_info_dir+0x10);
		}
#endif

#ifdef GEN_ZIP
		if( bAddZip )
		{
			static char pkdirend[]={
#if 1
			// WinRAR doesn't accept file as a valid archive without "PK34xxxx00xx" preceeding "PK56..."
			 0x50,0x4B,3,4, 0,0, 0,0, 0,0, 0,0,
			//ctrDirPtrSgn  thisDisk# cDirDisk  #cDirEntHere #cDirEntTotal cDirSize  cDirOff  commentSz
			 0x50,0x4B,5,6,    0,0,     0,0,         0,0,        0,0,       0,0,0,0, 0,0,0,0,  0,0 };
			int pkdirendSz = sizeof(pkdirend);
			memmove(buf+csize,pkdirend,pkdirendSz);
			*(long*)(buf+csize+pkdirendSz-06) = csize+pkdirendSz-22;
#else
			//pkFileEntSgn         v2extr gpBits method timeStm  CRC32    csize    usize   fnameLen extraFldsLen
			 0x50,0x4B,3,4,         10,0,  0,0,   0,0,  0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0,  2,0,   0,0, 'z','z', //+fileData
			//ctrDirEntSgn  vCreat v2extr gpBits method timeStm  CRC32    csize    usize   fnameLen extraFldsLen
			 0x50,0x4B,1,2,  22,0,  10,0,  0,0,   0,0,  0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0,  2,0,   0,0,
			// commentLen disk#start, intAttr, externAttr  filePtr
					  0,0,   0,0,       0,0,    0,0,0,0,   '?','?','?','?', 'z','z',
			//ctrDirPtrSgn  thisDisk# cDirDisk  #cDirEntHere #cDirEntTotal cDirSize  cDirOff  commentSz, optComment
			0x50,0x4B,5,6,   0,0,       0,0,         1,0,        1,0,        48,0,0,0, '?','?','?','?',0,0 };
			// timestamp (5:year1980,4:mon1,5:day1,5:hrs0,6:min0,5:sec0/2)

			int pkdirendSz = sizeof(pkdirend);
			memmove(buf+csize,pkdirend,pkdirendSz);
			*(long*)(buf+csize+pkdirendSz-28) = csize;
			*(long*)(buf+csize+pkdirendSz-06) = csize+32;
#endif
			csize+=pkdirendSz;
		}
#endif

		if( sect )
		{
// to fancy to use realLoadBase
//			long realLoadBase = (sect->file_addr&~0x1FF);
//			sect->virt_size = (((csize+255)&~255) + ((xsize+15)&~15) + 4 - realLoadBase + 0xFFF)&(~0xFFF);

			*(long*)(buf+nd32-4) = ((csize+0xFFF)&~0xFFF)-nd32; // where to decompress (reserve 4k or raw mode will GPF)

			sect->virt_size = (csize+xsize+0xFFF+8+0xFFF)&(~0xFFF);
			sect->file_size = csize - sect->file_addr;

			*(long*)(pe+0x50) = 0x1000 + sect->virt_size; // image size

			*(short*)(pe+0x5C) = subsystem;
		}

		if( buf[0]=='M' && buf[1]=='Z' && buf[12]==-1)
		{
		//	*(short*)(buf+02) = (csize%512);		// bytes in last page
			*(short*)(buf+04) = (csize+511)/512;	// #512 pages in file
			*(short*)(buf+14) = csize/16 + 2;		// initial stack offset
			*(short*)(buf+10) = xsize/16+0x1104;	// min addtional memory
		}
		flen = csize;
	}
}// infile
	if( outfile ) fwrite(buf,flen,1,fopenWB(outfile));
	char *buf3=0;
	if( bootfile )
	{
		int x = readFile(bootfile,&buf3);
	}
	if( FLPfile ) create_144_image(fopenWB(FLPfile),buf3,buf,plabel,flen);
#ifdef GEN_ISO
	if( ISOfile ) create_iso_image(fopenWB(ISOfile),buf3,buf,plabel,flen,ISOmode);
#endif
#ifdef GEN_FAT
	if( FATvol ) installBootSector(FATvol,buf3,plabel);
#endif

	//free(buf2);
	//free(buf);

	return 0;
}
/*
	{, label[12]
		int i;
		for( i=0; i<11 && plabel[i] ; i++ ) label[i]=toupper(plabel[i]);
		while( i<11 ) label[i++]=' ';
	}
*/
//					  "12345678ext"
//#define PART_EXE_NAME "PART    EXE"
static char tmpZZ[512*16];

int create_iso_image(OutStream ff, char* bootFat, char* bootExe, char *label, int bootExeSize,int ISOmode)
{
	int i, k, needZ=1, pos=0, nn;
	unsigned j;
	char *tmp=tmpZZ;

	int bootSect2k = (bootExeSize+0x7FF)&(~0x7FF);
	int totalSZ = 0x1C0000; // 1.44 floppy emulation
	if( ISOmode==0 )
		totalSZ = 0xD000 + bootSect2k;

	while(pos<totalSZ)
	{
		if( needZ )
		{
			memset(tmp,0,512*8);
			needZ=0;
		}
		nn = ((pos+0xFFF)&(~0xFFF))-pos;
		if( nn==0 ) nn=0x1000;
		//if( pos+nn>totalSZ ) nn=totalSZ-pos;
		if( nn>512*8 ) nn = 512*8;
		needZ++;

		const char *CD001 = "CD001\1""EL TORITO SPECIFICATION";

#define yyy 111
#define mmm 07
#define ddd 02
#define hhh 1
#define nnn 2
#define sss 0
#define gmt15 20
			// byte: length
			// byte: xarlen
			// dword*2: start
			// dword*2: datalen
			// bytes*7: y-1900,mm,dd,hh,mm,ss,gmt15offset
			// byte: file attributes (2-dir)
			// byte: interleaveSize, interleaveSkip
			//word*2: volSetSeqNum
		static char cRootDir[]={0x22,0,0x18,0,0,0,0,0,0,0x18,0,8,0,0,0,0, // 0x18 -> C000
					8,0,yyy,mmm,ddd,hhh,nnn,sss,gmt15,2,0,0,1,0,0,1,
					1,/*@21=0,@43=1*/1};

		if( pos==0x8000 ) // Primary Volume Descriptor (must reside in sector 0x10)
		{
 static char c78[]={01,00,00,01,01,00,00,01, // VolSetSize2x2, VolSetSeq2x2
 /*80*/00,0x8,0x8,00,0xA,00,00,00,00,00,00,0xA,0x14,00,00,00, // SectSize(2k)2x2, 84:PathTabLen4x2
 /*90*/00,00,00,00,00,00,00,0x16,00,00,00,00}; // 8C:LE PTab->A000, 94:BE PTab->B000, 9C:RootDir
			tmp[0]=1;
			memmove(tmp+1,CD001,6);
			memmove(tmp+0x78,c78,sizeof(c78));
			memmove(tmp+0x9C,cRootDir,0x21);
			//for(i=8;i<0x48;i++) tmp[i]=' ';
			//memmove(tmp+0x08,"LINUX",5); // System Identifier (32 bytes space-pad)
			//memmove(tmp+0x28,"CDROM",5); // Volume Identifier (32 bytes space-pad)
			*(long*)(tmp+0x50) = totalSZ/0x800; // Total number of sectors

			tmp[0x54]=tmp[0x53];
			tmp[0x55]=tmp[0x52];
			tmp[0x56]=tmp[0x51];
			tmp[0x57]=tmp[0x50];


			//for(i=0xBE;i<0x573;i++) tmp[i]=' ';
			//for(i=0x32D;i<0x370;i++) tmp[i]='0';
			//tmp[0x33D]=0xf0;
			//tmp[0x34E]=0xf0;
			//tmp[0x35F]=0;
		//	tmp[0x370]=0xF0;
		//	tmp[0x371]=1;
		//	tmp[0x372]=0;
			// ( pos==0x8800 )	// Boot Record Volume Descriptor (must reside in sector 0x11)
			strcpy(tmp+0x801,CD001);
			tmp[0x847]=0x19;	// -> C800 (BOOT.CAT)
		}
		else if( pos==0x9000 )
		{
			tmp[0]=255;
			memmove(tmp+1,CD001,6); // Volume Descriptor Set Terminator
		}
		else if( pos==0xA000 ) // LE PTab
		{
			tmp[0]=0x01;	// Length of directory identifier
			tmp[2]=0x18;	// 2-5: -> C000
			tmp[6]=0x01;	// 6-7: parent directory number
		}
		else if( pos==0xB000 ) // BE PTab
		{
			tmp[0]=0x01;
			tmp[5]=0x18;
			tmp[7]=0x01;
		}
		else if( pos==0xC000 )
		{
 static char c44[]={0x2c,0,0x19,0,0,0,0,0,0,0x19,0,8,
/*@50*/	0,0,0,0,8,0,yyy,mmm,ddd,hhh,nnn,sss,0xf0,0,0,0,
/*@60*/	1,0,0,1,
		0xa,0x42,0x4f,0x4f,0x54,0x2e,0x43,0x41,0x54,0x3b,0x31,0, // BOOT.CAT

/*@70*/	0x2C,0,0x1a,0,0,0,0,0,0,0x1a,0,0x80,0x16,0,0,0x16,
/*@80*/	0x80,0,yyy,mmm,ddd,hhh,nnn,sss,0xf0,0,0,0,1,0,0,1,
/*@90*/	0xa,0x42,0x4f,0x4f,0x54,0x2e,0x46,0x4C,0x50,0x3b,0x31,0}; // BOOT.FLP
			memmove(tmp+0x00,cRootDir,0x21);
			memmove(tmp+0x22,cRootDir,0x22);
			memmove(tmp+0x44,c44,sizeof(c44));

			// 0x8800 = BOOT.CAT

			// 0-1F, Validation Entry
			tmp[0x800]=1;
			//2: 0 - x86
			//4-1B: OEM ID String - anything
			tmp[0x81C]=0xAA; // 1C,1D Checksum word of 0-1F
			tmp[0x81D]=0x55;
			tmp[0x81E]=0x55;
			tmp[0x81F]=0xAA;
			// 20-3F - Initial Default Entry
			tmp[0x820]=0x88; //88 bootable
			tmp[0x821]=ISOmode; // 0 - No Emulation, 1-1.2, 2-1.44, 3-2.88, 4-HD80h
			// see int 13h, 4A,4B,4C
			// No emulation: retrieve status int134B01
			//2,3 - Load segment, 0 defaults to 7C0
			// 4 - System Type ?
			tmp[0x826]=1; // 6,7 - load this number of 512b sectors for the initial boot
			tmp[0x828]=0x1a; // 8:dword - Start address of the virtual disk -> D000

			if( ISOmode==0 )
			{
				*(long*)(tmp+0x7A) = bootExeSize;
				tmp[0x7E]=tmp[0x7D];
				tmp[0x7F]=tmp[0x7C];
				tmp[0x80]=tmp[0x7B];
				tmp[0x81]=tmp[0x7A];
				memmove(tmp+0x96,"EXE",3);
				*(short*)(tmp+0x826) = (bootExeSize+512-1) / 512;
			}



#undef sss
#undef nnn
#undef hhh
#undef mmm
#undef ddd
#undef yyy
#undef gmt15
		}
		else if( pos==0xD000 )
		{
			if( ISOmode==0 )
				pos+=WriteStream(ff,bootExe,bootExeSize);
			else
				pos+=create_144_image(ff,bootFat,bootExe,label,bootExeSize);
			continue;
		}
		else needZ--;

		WriteStream(ff,tmp,nn);
		pos+=nn;
	}
	return pos;
}


int create_144_image(OutStream ff, char* bootFat, char *bootExe, char *label, int bootExeSize)
{
	int i, k;
	unsigned j, resnn=0;
	char *tmp=tmpZZ;
	struct boot_fat16 *b=(struct boot_fat16*)tmp;

	unsigned exe_size = bootExeSize;
	unsigned num_sect = (exe_size+512-1) / 512;

	memmove(b,bootFat,512);
	memmove(b->label,label,11);
	unsigned part_num_sect=1 + 2*b->fat_size+b->root_entr/16+((b->fat_size*512*8)/12-2)*b->clust_size;
#if 0
	if( di!=0 && pPart!=0 )
	{
		if( part_num_sect > pPart->num_sect ) part_num_sect = pPart->num_sect;
		b->rel_sect = pPart->rel_sect;
		b->snum_sect = part_num_sect;
		b->num_sect = part_num_sect;

		b->drive_num=di->hd_num;
		b->media_desc=(di->hd_num==0?0xF0/*FD*/:0xF8/*HD*/);
		b->num_sects=di->sect_per_track;
		b->num_sides=di->num_heads;
	}
	//else Just use floppy defaults
	// b->ext_signat=0x29;
	// b->magic_num=MAGIC_NUM;
#endif
	resnn+=WriteStream(ff,tmp,512); // Boot sect

	memset(tmp,0,512*12);
	tmp[0]=0xF0; tmp[1]=tmp[2]=0xFF;
	for( j=0 ; j<num_sect ; j++ )
	{
		k=(j==num_sect-1?0xFFF:j+3);
		if( (j&1)==0 ) {tmp[j*3/2+3]=(k&0xFF); tmp[j*3/2+4]|=((k>>8)&0x0F); }
		else {  tmp[j*3/2+3]|=((k<<4)&0xF0); tmp[j*3/2+4]=(k>>4); }
	}
	resnn+=WriteStream(ff,tmp,512*9); // FAT1
	resnn+=WriteStream(ff,tmp,512*9); // FAT2

	memset(tmp,0,512*14);
	struct FatDirEntry *pDE = (struct FatDirEntry *)tmp;
	memmove(pDE->name_ext,label,11);
	pDE->cluster = 2;
	pDE->FileSize = exe_size;
	pDE->date=PM_DATE;
	pDE->time=PM_TIME;
	resnn+=WriteStream(ff,tmp,512*14); // Root

	int xx = 512*(num_sect-1);
	resnn+=WriteStream(ff,bootExe,xx); // Executable
	memmove(tmp+32,bootExe+xx,bootExeSize-xx);
	resnn+=WriteStream(ff,tmp+32,512); // Executable last sector

#if 0
	memmove(tmp,&part_exe_header,sizeof(part_exe_header));
	struct exe_header *mz=(struct exe_header *)tmp;
	unsigned init_ip;
	asm
	{
		push eax;
		mov	eax, 1*(startupEXE-offset0);
		mov init_ip, eax;
		pop	eax;
	}
	mz->n_bytes = exe_size%512;
	mz->n_pages = num_sect;
	mz->header_size = 2;
	mz->init_ip=init_ip;
	mz->min_add_par = ((unsigned)(&bss_end)+15)/16 - ((unsigned)(&bss_start)+15)/16 + 0x4000/16; //+stack
	mz->init_ss = ((unsigned)(&bss_end)+15)/16;
	char *p4k=(char*)(unsigned)4096;
	memmove(tmp+0x20,g_pq32env->code4k,4096);
	memmove(tmp+0x20+4096,p4k,512*14-4096-0x20);
	p4k+=512*14-4096-0x20;
	Sector nn=1+12+12+14;
	write_part(di,pPart,nn,14,tmp);
	nn+=14;

	for( j=14 ; j<num_sect ; )
	{
		memmove(tmp,p4k,512*8);
		int kk=8;
		if( j+8>=num_sect ) //last batch
		{
			kk=num_sect-j;
			memset(tmp+(kk-1)*512+exe_size%512,0,512);
		}
		j+=kk;
		write_part(di,pPart,nn,kk,tmp);
		nn+=kk;
		p4k+=512*kk;
	}

	if( pPart ) free_part_info(pPart);
	else
	{
		memset(tmp,'?',512*12);
		while(nn<2880)
		{
			int kk=min(2880-nn,12);
			write_part(di,pPart,nn,kk,tmp);
			nn+=kk;
		}
	}
#endif
	return resnn;
}



void installBootSector(char *FATvol, char*buf3, char* label)
{
	HANDLE h;
	DWORD temp;
	//DISK_GEOMETRY geo;
	char buf[512];

	//if( disk<2 )
//	sprintf(buf,"\\\\.\\%s",FATvol);
	//else
	//strcpy(buf,"\\\\?\\Volume{0e2b8bad-96e2-11e0-a42c-005056c00008}");
//	sprintf(buf,"\\\\.\\PHYSICALDRIVE%c",'0'+1);

	h = CreateFile(FATvol, GENERIC_READ | GENERIC_WRITE,
		FILE_SHARE_READ | FILE_SHARE_WRITE,
		NULL, OPEN_EXISTING, 0, NULL);

	if( h == INVALID_HANDLE_VALUE )
	{
		printf(cannotOpen,FATvol);
		return;
	}

	//BOOL i = DeviceIoControl(h,IOCTL_DISK_GET_DRIVE_GEOMETRY,NULL,0,
	//		&geo,sizeof(DISK_GEOMETRY),&temp,NULL);
	//if (i == FALSE) {
	//	CloseHandle(h);
	//	return;
	//}
//	printf("TracksPerCylinder = %d\n",geo.TracksPerCylinder);
//	printf("SectorsPerTrack = %d\n",geo.SectorsPerTrack);
//	printf("Cylinders = %d\n",geo.Cylinders.QuadPart);

	struct boot_fat16 bs, *pbs = (struct boot_fat16*)buf3;

	BOOL r = ReadFile(h,&bs,512,&temp,NULL);
	if( r!=TRUE || strncmp((char*)(bs.fs_id),"FAT1",4)!=0 )
	{
		printf("FAT1x not found on %s\n",FATvol);
	}
	else
	{
		memmove(bs.jmp,pbs->jmp,3);
		memmove(bs.label,label,11);
		memmove(bs.code,pbs->code,sizeof(bs.code));
		OVERLAPPED o;
		o.Offset=0;
		o.OffsetHigh=0;
		//DeviceIoControl(h,FSCTL_LOCK_VOLUME,NULL,0,NULL,0,&temp,NULL);
		int w = WriteFile(h,&bs,512L*1,&temp,&o);
	}


	CloseHandle(h);
	return;

}
