;
; Copyright (c) 1998-2011 Mikhail Ranish
;
; Permission to use, copy, modify, and/or distribute this software for any
; purpose with or without fee is hereby granted, provided that the above
; copyright notice and this permission notice appear in all copies.
;
; THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
; WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
; MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
; ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
; WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
; ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
; OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
;-----------------------------------------------------------------------------
;
; This stub will run under Windows 9x/XP/7, Linux, DOS, or boot under BIOS.
;
; Use pe2boot utility to compress relocatable PE and merge it with this stub.
; When *.exe is run this code will decompress and rebase attached executable.
; Then on Windows (or Wine) it will import symbols from linked external DLLs.
;
; When booting without OS or running under DOS:
;
;  - Memory is detected via int15 or XMS; A20 via int15, port 92, or AT KBC.
;  - CPU is switched to 32 bit mode, IRQs are handled and reflected to BIOS.
;  - 32-bit code can call real mode interrupt handlers with int86(REGS,REGS).
;  - Exceptions are trapped with debug monitor that supports single stepping.
;
;-----------------------------------------------------------------------------
; "Assembly language programming is an extravagant waste
;  of human talent and should be avoided whenever possible"
;  Peter Norton , 1983
;
;  This code was optimized for small size, and extreme fun optimizing it...
;
use16
;
; MZ and PE headers are fused (like TinyPE) but still work on all systems
;
	db "MZ" ;dec bp;pop dx;
	jmp short startBOOTP
;@2	dw (fileend) mod 512 ; number of bytes in last 512 page
	dw (fileend+511)/512 ; number of 512 byte pages in file
	dw 0; n_reloc_entries
	dw 0; header_size (paragraphs)
	dw 1100h; min_add_par
	dw -1	; max_add_par (no limit)
	dw (fileend+15)/16; init_ss;

pesig:
	dd "PE"				; PE signature				; dw init_sp, check_sum (0)
	dw 0x014C			; Machine (Intel 386)		; dw init_ip (0x14C = startMZ_15Ch - 10h)
	dw 1				; NumberOfSections 			; dw init_cs (1)
;@18:
start3c:
	pop	ss
	mov sp,16h
	pusha
	call start2b
offsetGDT:
gdt:
pe_addr:
	dd PEADDR_INIT
	dw csect-8-opthdr	; SizeOfOptionalHeader
	dw 0x10F;+0x2000 	; 100-32bit, 8-noDbgSym, 4-noLineNum, 2-Executable, 1-NoRelocs
			;+0x2000 - DLL / Entry point: DllMain(hmodule,reason,reserved)
opthdr:					; PE optional header
	dw 0x10B			; Magic (PE32)
	db "mr"	 ; my runtime signature
startBOOTP:
		mov	dl,3
start2:
		int	12h	; BIOS sets ax=free conv memory size in kb.
		shl	ax, 6	; now ax=top+1 paragraph of the free conv mem.
		mov cx, ss	; save if we need to access FAT BPB at [ss:bp]
		jmp short start1c
times (38h-$) db 'x'
	dd 0x1000+startPE-loadBase	; AddressOfEntryPoint
;@3C
	dd pesig			; MZ header: pointer to PE header
; LGDT and LIDT in 16-bit code use 24-bit addresses
pm_gdtr:	dw gdt_end-gdt-1
			dw GDTR32_INIT
times (0x44-$) db 'x'
	dd 0x400000			; PE ImageBase
	dd 0x1000			; SectionAlignment	; 1000h works on Win98, XP, and Wine
	dd 0x200			; FileAlignment		; Win XP needs multiple of 200h here
;@50
start2a:;dos
		mov dl, 4
start1c:;common
		mov al, 0
		sub ax,1100h
		db 0xEB ;jmp short +03
		dw 03	;@58	; MajorSubsystemVersion ; Windows version needs to be >= 3.10
		dw 10	;@5A	; MinorSubsystemVersion ;
		cld
		push ax;->ss
		jmp short start3c
;@60
pe_image_size:
	dd '????'			; SizeOfImage (must be aligned hdrsize + sum of all sections virtual sizes)
	dd 0x1000			; SizeOfHeaders - loader reads headers to PE_IMAGE_BASE and makes a ro page
rm_sp_ss:	dw -40h
rm_ss:		dw RM_SS__INIT
	dw 3;2				; Win32 Subsystem ; 2-GUI, 3-Console ; for more info search: msdn ms680339
	dw 0x8F00			; DllCharacteristics: 8000=TSaware,800=noBind,400=noSEH,200=noManifest,100=NXcompat,40-Relocatable
	dd 0x100000			; SizeOfStackReserve 1M ; Win32: initial esp = ~4Mb - reserve
	dd 0x1000			; SizeOfStackCommit  4k
	;dd 0x100000	;@78; SizeOfHeapReserve ; need to be some reasonable size ~?Mb
	;dd 0x1000		;@7C; SizeOfHeapCommit  4k
pm_idtr:	dw 78h*8-1
			dd IDTR32_INIT
	times (80h-$) db 0;'x'

N_EXC		= 11h	 ; Number of handled exceptions
TOTAL_INTS  = (N_EXC + 1 + 8)

NO_PUSH_NEXT = 10000000h

INIT_SEQ_ADDR  = (rm_ss-offsetGDT)
COPY_CODE_SIZE = INIT_SEQ_ADDR*16

RM_SS__INIT = (mov_esp+1-offsetGDT)
PM_ESP_INIT = (pm_idtr+2-offsetGDT	+NO_PUSH_NEXT)
IDTR32_INIT = (pe_addr  -offsetGDT	+NO_PUSH_NEXT) ; transition
PEADDR_INIT = (pm_gdtr+2-offsetGDT)

GDTR32_INIT = (pm32ea-offsetGDT)
PM32EA_INIT = (xchg_esp1+2 -offsetGDT)
XCHG_1_INIT = (mov_esi32+1 -offsetGDT)
PM_ESI_INIT = (code16_dsc+2-offsetGDT)
C16DSC_INIT = (code16_seg  -offsetGDT); + end seq flag (SF) already there

hiseg1:
		push dx;=0
		push si
		jmp short hiseg2
	times (84h-$) db 'x'
	;
	; Data directories (most entries are not used so storing GDT here)
	;
	dd 2				; NumberOfRvaAndSizes (data directories)
	dd 0, 0				; Export Table (0,?)
	dd idata+0x1000-loadBase; Import Table Addr must be >= 0x1000
	dd 16				; 	  Import Table Size: Win9x needs size>0, XP doesn't care
	;dd 0;@98,?			; Resource table; Win9x needs addr=0 regardless of #ddirs
	;dd ?,0 @BC			; Debugging information address and size (size must be 0)
	;dd 0 @D0,?			; Thread Local Storage (TLS) table (addr must be 0)
	;dd ?,0 @FC			; The CLR header address and size (if size!=0 linux loader assumes .net)


;idt_format:   offset  selector         DPL       offset
;			     0-15         Reserved P??0D110    16-31
;			    0000h, 0028h, 0 + 256* 10001110b,      0
;
;gdt_format:    limit  base    base cs:P??11CRA   GD00limit     base
;*3:unused       0-15  0-15   16-23 ds:P??10EWA   GD0016-19    24-31
;18:data32_dsc     0h,    0,  0 + 256* 10010110b, 11000000b + 256* 0 ; [4k..4G-1] base = 0 (0..4k-1 will GPF)
;20:code32_dsc 0ffffh,    0,  0 + 256* 10011010b, 11001111b + 256* 0 ; [0k..4G-1] base = 0
;28:code16_dsc 0ffffh,    *,  * + 256* 10011010b, 00000000b + 256* 0 ; [00..64k ] base = top 4k (under 640k)
;30:data16_dsc 0ffffh,    0,  0 + 256* 10010010b, 00000000b + 256* 0 ; [00..64k ] base = 0

data32_dsc:	dw		0,    0,  0 + 256* 96h,	0C0h    ; (over resource table addr=0)
code32_dsc:	dw	0FFFFh,   0,  0 + 256* 9Ah, 0CFh
code16_dsc:	dw	0FFFFh,C16DSC_INIT,16* 9Ah,  00h	; contains init seq SF but not CF
rm_idtr: ; Don't move: expecting this addr to be xx0h!!!
data16_dsc:	dw	0FFFFh,   0,  0 + 256* 92h,  00h
gdt_end:
	times (0xBC-$) db 0;'x'	; td32 doesn't like this to be non-zero
	dd 0				; Debug dir size @BC
; PE code section
csect:
csect_vs:
	dd '????'			; VirtualSize
	dd 0x1000			; VirtualAddress
	dd '????'			; SizeOfRawData (in file)
loadBase = 0x200
	dd loadBase    		; PointerToRawData (rounded on 0x200 down to 0), but not on ReactOS

	dd 0				; TLS addr @D0

kernel32:
	db "KERNEL32"
	times (csect+1Ch-$) db 0;
	dd 0xF0000000		; Characteristics: (x2=execute, x4=read, x8=write, x1=shareable)
hiseg2:
		push 10h	; default ext mem addr = 1Mb
		push dx;=0
		mov	si, XMS_call - offsetGDT
		add word [si-XMS_call+pm32ea], pm32_entry-offsetGDT
xms_chk:
		mov	ax, 4300h	; Check for XMS manager
		jmp short xms_chk2

		times (0xF0-$) db 'x'
LoadLibraryName:
	db "LoadLibraryA"
		times (0xFC-$) db 0
		dd 0; @FC ; unless CLR ddir size==0 linux loader tries to load mono
xms_chk2:
		int	2Fh
		cmp	al, 80h		; if XMS installed al==80
		mov	al, 10h		; ax=4310h ; Get XMS driver entry point in es:bx
		je  xms_init	; if XMS installed al==80
;a20_init:
		mov	ax, 2401h	; Enable A20 gate (2400h to disable)
		int	15h 		; First, try BIOS service (mods ah)
		jnc	a20_int15ok

		cli	; disable interrupts while working with ports and protect stack

		mov	dl, 92h	; Then, try PS/2 system control port A (92h)
		in	al, dx  ; bit 1 - A20 line, bit 0 - reboot
		btc	ax, 1	; set bit 1 to enable a20 (clear to disable)
		jnc	a20_out	; only output to port if bit2 isn't alreday set

		mov	dl, 64h	; Finally, try AT keyboard controller
a20_wait:
		in	al, dx	; Read KBC status
		test	al,2; if bit 1 is clear buffer is ready for write
		loopnz	a20_wait
		mov	al, 0DFh; a20 control command: DF - enable, DD - disable
a20_out:
		out	dx, al
		;sti - no need for sti, will cli soon
a20_int15ok:

; Get memory information (high words of EAX and EDX have been set to 0 before)

		mov	ax, 0E801h
		int	15h
		jc	mem_use1588
		; AX = extended memory between 1M and 16M, in K (max 3C00h = 15MB)
		; BX = extended memory above 16M, in 64K blocks
		; CX = configured memory 1M to 16M, in K
		; DX = configured memory above 16M, in 64K blocks
		; some BIOSes return AX=BX=0
		; first try AX+BX*64, then try CX+DX*64
mem_shl6:
		xchg	bx, dx
		cmp	ax, 3C00h		; is there full 15 mb?
		ja		mem_use1588	; AH=86h unsupported function, AH=80h invalid command, or E8 still in AH
		je		mem_15m		; there is full 15mb and no ISA hole below 16mb
		cwd		;dx=0		; memory hole - ignore any memory above 16mb - we want continuous memory
mem_15m:
		shl	edx, 6			; 64k blocks -> kbytes
		add	edx, eax
		jnz	mem_ok			; got memory size in edx
		add	ax, cx
		jnz	mem_shl6
mem_use1588:
		mov	ah, 88h			; Get amount of memory above 1M in K (ax)
		int	15h				; CF set if error
mem_xchg:
		xchg	ax, dx		; now edx = mem size (k), mem base addr on stack
        jnc	mem_ok
mem_error:
        cdq ; if error set edx=ext_mem_size to 0
mem_ok:
        pop edi ; edi=ext_mem_addr, edx=ext_mem_size (k) (0 if mem or a20 error)
		jmp short switch_mode_pushad


XMS_cleanup:
			db	0BAh	; mov dx, xms_handle
xms_handle:	dw	1
			dec		dx
			jz	XMS_call_ret

			mov	ah, 0Dh		; Unlock XMS memory block dx (mods ax,bl)
			call	XMS_call
			mov	ah, 0Ah		; Free XMS memory block dx (mods ax,bl)
			times 15Ch-1-$ db 90h
XMS_call:
			db	9Ah ;call far XMS_GATE; returns: AX=1 success, AX=0 failure
XMS_addr:
startMZ15C:	; Entry point under DOS (+1:14Ch == +0:15C)
				mov ax,[2]	; psp:[2] = top seg of free conv memory
				db 6Ah		; push ib=40h -> jmp +1:40h == +0:50h
			;or ax, ax		; jz error
				inc ax		; opcode=40h
				dec ax		; jz error
XMS_call_ret:
			ret

xms_init:
		;mov	al, 10h		; ax=4310h ; Get XMS driver entry point in es:bx
		int	2Fh
		mov	[si+(XMS_addr-XMS_call)+0], bx
		mov	[si+(XMS_addr-XMS_call)+2], es
		mov	ah, 05h		; Local enable A20 line
		call	si;XMS_call
		jz		short mem_error ; no_mem
		;ah=00h	already	; Get XMS version
		call	si;XMS_call	; AH=Major, AL=Minor, mods bx,dx
		add	ah, -3		; Version 3+ supports >64M -> CF
		mov	ah, 8*2		;
		rcr	ah, 1		; if(XMS 1,2) ah=8 ;; if(XMS 3+) ah=88h

		push	ax		; 08h or 88h; Query extended memory
		xor		bx, bx	; bx=0
		call	si;XMS_call	; (e)ax = largest block (k), (e)dx = total, mods ecx
		dec 	bx		; if bl!=0 error
		jns		short mem_error
		xchg	eax, edx
		pop 	ax
		push	dx		; save mem size (lo)

		inc 	ah		; 09h or 89h; Allocate memory block (e)dx (k)
		call	si;XMS_call	; returns dx = handle (null if no memory allocated)
		jz		short mem_error

		mov	ah, 0Ch		; Lock extended memory block dx=handle
		add	[si+(xms_handle-XMS_call)], dx
		call	si;XMS_call	; returns: dx:bx 32-bit physical address (CF=0)
		jz		short mem_error
		pop	ax	; mem size (lo)
		push	dx  ; mem_addr+2
		push	bx  ; mem_addr+0
		jmp 	short mem_xchg ; nc

;
; Switch PM->RM
;
;	Disable interrupts
;	Disable paging and transfer control to code with linear == physical address (N/A)
;	Load ds,es,ss with a 64k writable (W=1) byte granular (G=0) and expand up (E=0) valid (P=1) selectors
;	Jump to 64k 16-bit code selector
;	Restore real mode IDT
;	Clear PE bit in CR0
;	Jump to a real mode segment
;	Load segment registers with valid real mode segment addresses
;
;
; Switch RM->PM
;
;	Disable interrupts
;	Load GDT and IDT
;	Set PE bit in CR0
;	Jump to a protected mode selector
;	Load LDT (N/A)
;	Load task register (N/A)
;	Load valid protected mode selectors
;

call_rm_int:
		popad
		; original flags are on stack ; if they have sti iret will enable ints
		; so PM handlers could miss some IRQs - still rm handlers will get them
		call far[-40h+4*(8+4)] ; ds=es=rm_ss ; call (ebx=rm_int_addr)
		;mov sp,-22h ; if callee retf instead of iret this will fix it
		pushf;only need low 16 bits, high 16 bits are on stack
switch_mode_pushad:
		pushad
switch_mode:
		cli
		mov bx, data16_dsc-gdt ; 30h (mod 16 -> bit 0)
		mov esp, cr0
		btc sp,bx ; PE->CF ;; flip PE ;; other flags are UNDEFINED!
		jc	pm2rm ; PE=1 ;; need to switch [protected -> real mode]
		;jnc rm2pm ; PE=0 ;; need to switch [real mode -> protected]
rm2pm:;CF=0
		lgdt	[cs:bx-(data16_dsc-gdt)+pm_gdtr-offsetGDT]
		lidt	[cs:bx-(data16_dsc-gdt)+pm_idtr-offsetGDT]
		mov bl, data32_dsc-gdt
		mov cr0, esp
			db	66h,0EAh;jmp far to Protected Mode
pm32ea:		dd	PM32EA_INIT	;jmp to pm32_entry
			dw	code32_dsc - gdt

pm2rm:;CF=1
		mov ds, bx
		mov es, bx
		mov ss, bx

		lidt [cs:bx-(data16_dsc-gdt)+(rm_idtr-offsetGDT)]

		mov cr0, esp
			db	0EAh;jmp to Real Mode
			dw	rm16_trg - offsetGDT
code16_seg:	dw	C16SEG_INIT
rm16_trg:
		lss sp, [cs:bx-(data16_dsc-gdt)+rm_sp_ss-offsetGDT]
		push ss
		push ss
		pop es
		pop ds
		jcxz call_rm_int; cx==0 for rm_int ; cx!=0 for app exit

;exit_sti:
		sti
		call XMS_cleanup
exit_mov:
		xchg	ax, si	; get exit code to al
		mov 	ah, 4Ch	; application exit code passed in al
		int 	21h		; if there is no OS this will return
		int 	18h		; tell BIOS to try next boot device
		;;;int 	19h		; BIOS reboot function
		;hlt
exit_jmp:
		jmp short exit_mov
		;
		;  Check that CPU is 386+
		;
;		push	sp	; 80386+ first push then decrement
;		pop 	bp	; 8086-286 first decrement then push
;		cmp	bp, sp	; if poped bp < sp it is not a 80386
;		jnc	its386, je
;		jc	not386, jne

start2b:
		pop si;->offsetGDT

		push	ds	; save PSP @ss:04

		add ax, 1000h	; +64k

		mov es, ax
		mov ds, ax

		mov cx, COPY_CODE_SIZE

		jmp short start2c

		times 200h-2-$ db (90h)
		dw 0xAA55		; magic for BOOTP
startBIOS:; @200h
		; Boot loader entry point => cs:200h, dl=disk# ss:bp->BPB
		; dl=drive# {A:0,C:80h} dh:bit5=0 if there is int 13h support
		; Initial ss:sp is about ~8k from the top of int 12h memory
		jmp start2
start2c:

		mov al, COPY_CODE_SIZE/16

		xor di, di		; ns, nc

		rep cs movsb    ; copy own code

		push ds			; gdtr, esi32_init, pm32ea, code16_dsc, code16_seg
		push cs			; pe load addr
		push ax			; idtr
		push ss			; rm_ss, pm_esp

		cbw	; ax -> rm_ss
		; need ns, nc coming here
seg2addr32:
		shl	dword [bx], cl ; mods CF, SF, ZF (if cx=0 does nothing)
		xchg ax, bx
		pop  ax
		jnc	save_seg
		db 0xB9, 4 ; mov cx, {push ax}4
save_seg:
		push ax
		xchg ax, [bx]
		cmc
		jns seg2addr32

init_idt:
;		mov	bx, ((int_handler-offsetGDT)-(TOTAL_INTS)*4)
C16SEG_INIT	 =  ((int_handler-offsetGDT)-(TOTAL_INTS)*4)
		xchg ax, bx
init_idt_ii:
		; [ds:bx] <= {push int#=di/8; jmp int_handler}; 6A,#,EB,?
		; [es:di] <= {addr0-15, code32_dsc, 8E00h, addr16-31}
		mov	dword [bx], (((int_handler-offsetGDT) mod 100h )*100h + 0xEB)*10000h + (5003h - COPY_CODE_SIZE)
		add	[bx], di
		rol word [bx],5
		lea	eax, [bx+4]
		xchg	ax, bx

		mov		dx, es
		rol		dx, 4
		or		ah, dh	;  ax=addr0-15, dl=addr16-23, dh will be cleared

		stosw
		cmp	al, ( (int_handler-offsetGDT)-(TOTAL_INTS)*4 + (N_EXC)*4 - 4) mod 100h
		mov	ax, code32_dsc-gdt
		stosw
		mov	ax, 8E00h
		mov	cx, (70h-1-N_EXC)*8/4+1
idtz:
		stosw
		xchg ax, dx
		cbw
		stosw
		cdq 	; edx=0 (EDX<-sign EAX), high bits of eax cleared by "lea eax," above
		loope	idtz
		sub	[bx-1], bl
		jnz	init_idt_ii

		smsw	bp;=cr0	; smsw is unprivileged and should not cause GPF if CPU is in PM
        shr 	bp, 1		; CF=PE protected mode bit

		jc  	exit_jmp	; CPU is already in protected mode

		push hiseg1-offsetGDT; cs already on stack
		retf

use32
int_handler:
		;
		; Interrupts:
		;
		;  0h-07h -> exc
		;  8h-0Fh -> exc or irq
		; 10h-17h -> exc
		;     6Fh -> rm_int
		; 70h-77h -> irq
		;
		push	ds
		push	es
		pushad
		cld
		mov ebp, esp
		push ss
		pop eax
		mov ds, ax
		xchg eax, [ebp+4*(8+2)] ; int#
		mov  edi, [ebp+4*(8+2+1+2)] ; orig eflags
		cdq	;edx=0
		cmp al, 6fh
		je int_6fh	; int 6fh -> rm_int
		xchg eax, ecx
		mov al, 0Bh
		mov dl, 20h
		ja int_irq	; it is irq 70-77h
		test cl, 8
		jz exc_dump	; must be an exception 0-7 or 10h-17h
		; could be either exc or irq 8-F
		out	dx, al	; Read Interrupt Request Register from 8259A
		in	al, dx
		mov	ah, al	; move status word to bits 8-Fh of eax
		bt eax, ecx	; eax[bit:cl] => CF
		jc int_irq	; cf -> pending IRQ
exc_dump:
		call exc_dump2
ERRCD EQU 194h/2 ; Exception has extra error code pushed on the stack
NOERR EQU 178h/2 ; no error code
exc_msg1:
	DB	NOERR ;	,"DIV"		;00 - Divide by 0 (*)
	DB	NOERR	,"DBG"		;01 - Debug (*)
	DB	NOERR ;	,"NMI"		;02
	DB	NOERR ;	,"INT3"		;03 - Breakpoint
	DB	NOERR ;	,"INTO"		;04 - Overflow
	DB	NOERR ;	,"BOUND"	;05
	DB	NOERR ;	,"UD"		;06 - Undefined instruction
	DB	NOERR ;	,""  		;07 - Obsolete FPU exception
	DB	ERRCD ;	,"DF"		;08 - Double Fault
	DB	NOERR ;	,""   		;09 - CoProcessor Overrun (Obsolete)
	DB	ERRCD ;	,"TSS"		;0A - Invalid TSS
	DB	ERRCD ;	,"NP"		;0B - Segment not present
	DB	ERRCD ;	,"SS"		;0C - Stack Fault
	DB	ERRCD	,"GPF"		;0D - General Protection Fault
	DB	ERRCD ;	,"PAGE"		;0E - Page Fault
	DB	NOERR ;	,""  		;0F - Intel Reserved
	DB	NOERR ;	,"MF"		;10h - x86 Floating Point
	;DB	ERRCD ;	,"AC"		;11h - Alignment Check
	;DB	NOERR ;	,"MC"		;12h - Machine Check
	;DB	NOERR ;	,"XF"		;13h - SIMD
	;DB NOERR ; end of list indicator (using retf opcode instead)
	;N_EXC EQU 11h	 ; Number of handled exceptions

popf_retf8:
		popfd
		retf 8	; skip eflags and rm_int#

int_6fh:
		mov ecx, [ebp+4*(8+2+1+3)] ; rm_int#
int_irq:
		cli	; must have exclusive access to rm_stack area and int table
		xor eax, eax ; nc,z
		mov ebx, [cs:4*ecx+eax] ; ebx = rm_int addr
		;nc,z
		db 0beh ; mov esi, <next 4 bytes> ; skip next 4 bytes
xchg_nc_nz:
		db 0ch ; or al,<ib> (nc,nz) ; es: ; skip next 2 bytes
xchg_esp_esi:
		xchg esp, [esi]
		popad
		cmc
xchg_esp1:
		xchg esp, [XCHG_1_INIT]
		ja ret2u ; nc & nz
		pushad
		jnc jmp_pm16 ; nc & z ; nc sets ecx=0

pm32_entry:; nc, bx=data32_sel, edi=ext_mem_addr, edx=ext_mem_size (k) (0 if error)
mov_esi32:
		mov	esi, PM_ESI_INIT
		jc xchg_esp_esi ; not_init, looping through the stack exchange
		mov ds, bx
		mov es, bx
		mov ss, bx
		lea esp, [esi-40h]
		mov ecx, [esi-4] ; on init = 1Mb, 0 if reentering from rm_int
		jecxz xchg_nc_nz

		;check A20 line ; assuming cli, edx!=0
		xor 	edx, [esp-4+ecx]; now edx!=[esp-4+1Mb]
		push	edx				; push to  [esp-4]
		xor 	edx, [esp+ecx]	; if A20 doesn't work [esp]==edx==[esp+1Mb] => edx=0, zf
		;jz	A20_mem_problem
		;jnz A20works
mov_esp:mov esp, PM_ESP_INIT
		mov ebp, esi
		xchg esi, [esi] ; esi=pe_load_seg, PM_ESI -> [esi]
		sti				; after A20 check and esi init can enable interrupts

		shl 	edx, 10 ; ext mem at eax, edx=size -> convert kb to bytes

		pop ebx ; codeend relative to pe_addr
		lea esi, [esi+ebx-(COPY_CODE_SIZE+offsetGDT)+codeend]; -> pe_addr+codeend
		cs		; data32 has first 4k page protected
		lodsd	; decomp_dest_offset
		add eax, esi	; decompressed code destination

		mov ecx, esp	; free_conv_size = (top-~4k) - 64k buffer
		sub ecx, eax	; free conv mem at eax, ecx = size
		jb	exit2rm		; mem overrun: stack below end of compressed code
		;int 1
		pushad	; save memory info
		;eax ; low_mem_addr ; conv mem below 640k
		;ecx ; low_mem_size at eax
		;edx ; ext_mem_size at edi (or 0 if err)
		;ebx ; codeend relative to pe_addr
		;esp ; buffer_addr+4
		;ebp ; code16_addr
		;esi ; pe_addr+codeend+4
		;edi ; ext_mem_addr if edx!=0
		cmp [ebp-offsetGDT+pe_image_size], edx ;!!! should be decomp_size
		jb	call32		; use extended memory
		cmp [ebp-offsetGDT+pe_image_size], ecx ;!!! should be decomp_size
		jae exit2rm		; neither conv nor extended memory is large enough
		xchg eax, edi	; use conv memory
call32: ; need cf coming here
		sbb ebp, ebp	; ebp=-1 to indicate it is not win32
		call call2decompress32
exit32: ; void exit32(int exit_code)
		popad		; esi=exit_code
exit2rm:
		stc	; to exit
jmp_pm16:
		sbb ecx, ecx	; NC -> ecx=0 for rm_int ; C -> ecx=-1 for exit
		db	66h,0EAh	; jump from code32 to code16 selector
		dw	switch_mode - offsetGDT
		dw	code16_dsc - gdt

ret2u:
		xchg [ebp+4*(8+2)], edi ; returned eflags
popad_iretd: ; if edx==0 it was int 6f
		dec edx
		popad
		pop es
		pop ds
		js popf_retf8
		add esp,4 ; skip ss/eflags
		iretd

SBBAL EQU 68h ; to access bits must be 68h or less

exc_msgret:
		pop	esi
		push	eax
exc_lodsb:
		xor eax, eax
		cdq	;edx=0
		lodsb
		cmp	cl, al
		sbb	al, SBBAL
		jbe	exc_nolea
		lea	ebx, [esp+4*eax+4*(SBBAL+8+1+8+2+1+3+1-80h)]
exc_nolea:
		jl	exc_test_char
		mov	ebx, [cs:ebx]	; cs: to lookup code below 4k
		mov edi, ebx		; save eflags
		;	  ODI SZ   P C
		; edi=111011000101b	; eflags test pattern
exc_test_bits:
		xchg eax, ecx
		rol edi, cl
		xchg eax, ecx
		mov	al, 0
		js	exc_adc_al ; print bit
exc_wr_reg:
		mov	dl, 8
exc_next_nib:
		mov	al, 0Fh
		rol	ebx, 4
		and	eax, ebx
exc_adc_al:
		adc	al, '0'
		cmp	al, '9'+1
		jc	exc_wr_char
exc_wr_char7:
		adc	al, 7h
exc_wr_char:
		mov	ah, 0Eh	; output char
		push 10h ; rm_int 10h
		int	 6fh
		dec	edx
		js	exc_lodsb
		jnz	exc_next_nib
		mov	al, ' '-SBBAL
exc_test_char:
		add	al, SBBAL-8
		jo	exc_esi
		jnc	exc_test_bits ; jns write_reg ; js bits
		jnz	exc_wr_char7
exc_done:
		popad
		pop	ebp
		loop	exit2rm ; quit if exception is not DEBUG (ecx!=1)
exc_ask_again:
		push 16h ; rm_int 16h (ah=0) get a key
		int	 6fh
		and	 eax, 5Fh	; turn to uppercase and clear ah

		cmp	al, 'Q'
		je short exit2rm

EFLAGS1ADDR	EQU	[ebp+4*(8+2+1+2)+1] ; eflags+1 for int 1 (no err)

		btr	EFLAGS1ADDR, ecx ; clear trap bit

		cmp	al, 'C'
		je	short popad_iretd

		bts	EFLAGS1ADDR, ecx ; set trap bit
		cmp	al, 'S'
		je short popad_iretd
		jmp	short exc_ask_again

exc_dump2:	; ecx=exc#, ds=ss, bases of all (ds,ss,cs)==0
		pop esi ; esi=exc_msg1
		push ebp
		push ecx
		push ecx
exc_next0:	; find exception name
		lodsb
		shl	al, 1
		sbb	cl, ch;=0
		jns	exc_next0

		add	cl, al

exc_esi:
		xchg	eax, esi
		jnp	exc_msgret
		call exc_msgret

RR EQU 76h
RF EQU 0E9h ; 80h+SBBAL+1 + (68h-SBBAL)
SWNL  EQU 192h/2

;exc_msg2:
	db 13, 10
	; rr-12=exc#, RR+3=cs, rr+2=ip, rr+1=ip/errc
	db "##", RR-12, RR+3, RR+1, ":", 80h, SBBAL, ":", 86h, SBBAL, 13
	;db "      ", RR+3, RR+1, 13, RR-12;, 13 ; optional
	db NOERR, ' '
	db SWNL, "AX=", RR-2, "BP=", RR-7
	db SWNL, "BX=", RR-4, "SI=", RR-7
	db SWNL, "CX=", RR-1, "DI=", RR-7
	db SWNL, "DX=", RR-1, RR+8,  RF+1, RF+1, RF+29-256, RF+4, RF+1, RF+4, RF+2
	db SWNL, "IP=", RR+7, SBBAL,"DIOSZPC" ; @10, @9, @11, @7, @6, @2, @0
	db 13, 10 ;,"C,S,Q"
	db "?", 8


startPE2:
		pop ebp
		lea esi, [ebp-stPEret+codeend]
		lodsd ; decomp_dest_offset
		lea edi, [esi+eax] ; decompressed code destination
decompress32:

		; esi=src, edi=dst_base

		xor	eax, eax; nc
		inc	eax
		push eax	; 1
		push eax	; 1
		mov ebx, esp

load_ecx: ;cf
		xchg eax, ecx
		cs			; use cs becuse access to lower 4k will GPF
		lodsd		; load next dword from control stream
		jc get_bit
		xchg eax, edx
		; after pop eax=1, ecx=1, edx=initial control dword
dsave:
		pop eax
		push edi
		push edi
ddown:
		xchg edi, [esp+4*ecx]
		loop ddown
dcopy:
		xchg esi, [esp]
		rep movsb	; if ecx==0 rep does nothing
		pop esi
dmove:
		cs			; use cs becuse access to lower 4k will GPF
		rep movsb

next_len:
		lea ecx, [edx-1]	; control word
		xor ch, ch	; nc
		; must come here with nc
prefix_bit:
		inc cl				; range -pp..0
		ja get_bit			; nz & nc
		sbb cl, dl			; shift range to 0..pp-1,pp
		jnc get_bit
		add cl, dh
		shr cl, cl
		shr ecx, cl
get_bit:
		xchg eax, ecx
		rcr ecx, 1			; get bit from control stream
		jecxz load_ecx
		xchg eax, ecx
		js  prefix_bit
		rcl ecx, 1			; copy fixed bits into result
		jnc get_bit
op2_bit:					; after move [ebx]==dl else 0
		inc ecx
		xor [ebx+5], dl		; byte[ebx+5] is guaranteed to be 0 initially
		rcr ecx, 1
		ja  dmove	; nc & nz	; [ebx]!=0
		jnz op2_bit	;  c & nz
		inc ecx
		jnc dcopy	; nc &(z)	; [ebx]==0
					;  c &(z)	; [ebx]==0
		push eax
		lea eax,[esp+4*ecx]
		sbb eax, ebx
		sar eax, 3
		js dsave	; s, nz, c/nc
		dec eax
		jnc	dint32
;base:
		add eax, [ebx]	; compute base+offset ; nc,nz,ns ; if eax was -1 -> cf -> exit
dint32:
		stosd
		pop eax
		jnc next_len	; nc & nz after base

		; DLL imports on Win32

		pop esi ; dll import list
imp_next_dll:
		lodsd			; next dll name
		cmp eax, ebp	; ebp = -1 when not on Win32
		jb imp_done		; not window or end of list
		push eax
		lodsd
		xchg eax, edi ; dll ttab addr
		call dword [ebp-stPEret + LoadLibraryFptr] ; LoadLibrary ; stdcall ret 4
		pushad
		xchg eax, ebx ; ebx=hmodule (dll base)
		mov esi, [ebx+0x3C] ; pe offset
		mov esi, [ebx+esi+0x78] ; export dir rva
		lea esi, [ebx+esi+0x18] ; export dir addr + 0x18
		lodsd	; +18 ; count
		xchg eax, ecx ; count
		lodsd		; +1Ch ; funcs[] rva
		lea edx, [ebx+eax] ; funcs[] addr
		lodsd		; +20h ; names rva
		mov ebp, [esi];+24 ; ordinals rva
		lea esi, [ebx+eax] ; names addr
imp_next_name:
		lodsd	; eax = exported function name
		pushad
		; h=5381;=1505h; compute djb2 hash
		; do{ h=(h*33)^(*p); } while(*p++);
		lea esi, [ebx+eax]; dll base + func name
		mov ch,15h	; ecx=5381=1505h
		mov cl,05h
hash_char:
		imul ecx, ecx, 33
		lodsb
		xor cl, al
		das	; ZF will be set only if al==0
		jne hash_char
		xchg eax, ecx	; eax = hash
ttab_next:
		mov ecx, [edi]
		jecxz ttab_end
		scasd           ; hash - *ttab
		jne ttab_next
		;get function ptr
        movzx eax, word [ebx+ebp]; func ordinal
		add ebx, [edx+4*eax]  	 ; func addr + base
		mov [edi-4], ebx
ttab_end:
		popad
		inc ebp ; next ordinal
		inc ebp
		loop imp_next_name
		popad
		jmp short imp_next_dll

idata:	; Import Table fields:
	dd 0; HintTable		; Win 9x and ReactOS need 0
	;dd 0; TimeDateStamp ; UNUSED ; Win9x owerwrites this. Ouch!!!
	;dd 0; ForwarderChain; UNUSED ; Win9x owerwrites with HMODULE -> KERNEL32
call2decompress32:
		call decompress32
		;[esp+4]=eax ; not enough space to implement
		ret ; -> exit32
		dw 0
	dd kernel32						;	DllName
	dd kernel32_iat+0x1000-loadBase	;	ThunkTable
startPE:
	call startPE2
stPEret:
imp_done:
		ret 8
kernel32_iat:
LoadLibraryFptr:
	dd LoadLibraryName - 2
	dd 0							; DllName=0 end of list
codeend:
fileend:
