#ifndef PE_STUB_H
#define PE_STUB_H

#define GET_BOOT_DEVICE(pe_stub_env)  ((pe_stub_env)->buffer[0x10])
#define EXIT32(pe_stub_env,exit_code) (((void (*)(int))((pe_stub_env)->exit32_addr))(exit_code))

#define BOOT_HD0 0x80	// HardDisk 0
#define BOOT_HD1 0x81
#define BOOT_FDA 0   	// Floppy/USB 0
#define BOOT_FDB 1
#define BOOT_PXE 3   	// BOOTP/PXE
#define BOOT_DOS 4   	// MS-DOS, FreeDOS, ...

#pragma pack(push,1)
struct pe_stub_env
{
	unsigned
		exit32_addr,	// pointer to exit function for non-Win32 env // exit code for DOS
		ext_mem_addr,	// edi;
		__esi,			// esi; compressed stream src addr
		code16_addr,	// ebp;
		__esp_buf_4, 	// esp; -> buffer[4]
		__ebx,			// ebx; codeend relative to pe_addr
		ext_mem_size,	// edx; or 0 if A20/XMS error
		low_mem_size;	// ecx; low_mem_addr..buffer
	union{
	unsigned
		low_mem_addr;	// eax; conv mem below 640k
	char
		buffer[0x10000];// 64k // ?, ?, ds, di, si, bp, sp, bx, dx, cx=init_ss, ax=ss=buffer_seg
		// init_ds @04h	// PSP segment under DOS, command line is at PSP*10h+80h
		// init_dx @10h	// boot disk# FDA:0, HD0:80h (dh:bit5=0 if int 13h supported), PXE:??03, DOS:??04
		// init_ss @12h	// init_ss
	};
	char
		code16[0x1000];	// <- code16_addr
		// gdt@0..97h
		// code .......
		// idt @4A0h,+8*78h
		// unintialised..4k
};
#pragma pack(pop)

#define EFLAGS_C 0x01
#define EFLAGS_P 0x04
#define EFLAGS_Z 0x40
#define EFLAGS_S 0x80

#pragma pack(push,1)
typedef union REGS
{
	struct{unsigned long  eax,   ebx,   ecx,   edx,   esi,   edi,   ebp,  eflags; };
	struct{unsigned long  eax,   ebx,   ecx,   edx,   esi,   edi,   ebp,  eflags; } d;
	struct{unsigned short ax,EA, bx,EB, cx,EC, dx,ED, si,ESi,di,EDi,bp,EBp,flags; };
	struct{unsigned short ax,EA, bx,EB, cx,EC, dx,ED, si,ESi,di,EDi,bp,EBp,flags; } x;
	struct{unsigned char  al,ah, a2,a3, bl,bh, b2,b3, cl,ch, c2,c3, dl,dh, d2,d3; };
	struct{unsigned char  al,ah, a2,a3, bl,bh, b2,b3, cl,ch, c2,c3, dl,dh, d2,d3; } h;
}REGS;
#pragma pack(pop)


#ifdef  __cplusplus
extern "C" {
#endif

int int86(int rm_int, REGS *regs_in, REGS *regs_out);

//int _inp(unsigned port);
//int _inpw(unsigned port);
//int _inpl(unsigned port);
//int _outp(unsigned port, unsigned value)
//int _outpw(unsigned port, unsigned value)
//int _outpl(unsigned port, unsigned value)

void exit(int c);


#ifdef  __cplusplus
};
#endif

#endif
